# frozen_string_literal: true

# Released under the MIT License.
# Copyright, 2020-2024, by Samuel Williams.

require 'samovar'

require_relative '../general'

require 'console/terminal'

module Process
	module Metrics
		module Command
			module Bar
				BLOCK = [
					" ",
					"▏",
					"▎",
					"▍",
					"▌",
					"▋",
					"▊",
					"▉",
					"█",
				]
				
				def self.format(value, width)
					blocks = width * value
					full_blocks = blocks.floor
					partial_block = ((blocks - full_blocks) * BLOCK.size).floor
					
					if partial_block.zero?
						BLOCK.last * full_blocks
					else
						"#{BLOCK.last * full_blocks}#{BLOCK[partial_block]}"
					end.ljust(width)
				end
			end
			
			class Summary < Samovar::Command
				self.description = "Display a summary of memory usage statistics."
				
				options do
					option '--pid <integer>', "Report on a single process id.", type: Integer, required: true
					option '-p/--ppid <integer>', "Report on all children of this process id.", type: Integer, required: true
					
					option '--memory-scale <integer>', "Scale maximum memory usage to the specified amount (MiB).", type: Integer, default: 512
				end
				
				def terminal
					terminal = Console::Terminal.for($stdout)
					
					# terminal[:pid] = terminal.style(:blue)
					terminal[:command] = terminal.style(nil, nil, :bold)
					terminal[:key] = terminal.style(:cyan)
					
					terminal[:low] = terminal.style(:green)
					terminal[:medium] = terminal.style(:yellow)
					terminal[:high] = terminal.style(:red)
					
					return terminal
				end
				
				def format_processor_utilization(value, terminal)
					if value > 80.0
						intensity = :high
					elsif value > 50.0
						intensity = :medium
					else
						intensity = :low
					end
					
					formatted = "%5.1f%% " % value
					
					terminal.print(formatted.rjust(10), intensity, "[", Bar.format(value / 100.0, 60), "]", :reset)
				end
				
				UNITS = ["KiB", "MiB", "GiB"]
				
				def format_size(value, units: UNITS)
					unit = 0
					
					while value > 1024.0 && unit < units.size
						value /= 1024.0
						unit += 1
					end
					
					return "#{value.round(unit)}#{units[unit]}"
				end
				
				def format_memory_usage(value, terminal, scale: @options[:memory_scale])
					if value > (1024.0 * scale * 0.8)
						intensity = :high
					elsif value > (1024.0 * scale * 0.5)
						intensity = :medium
					else
						intensity = :low
					end
					
					formatted = (format_size(value) + ' ').rjust(10)
					
					terminal.print(formatted, intensity, "[", Bar.format(value / (1024.0 * scale), 60), "]", :reset)
				end
				
				def call
					terminal = self.terminal
					
					summary = Process::Metrics::General.capture(pid: @options[:pid], ppid: @options[:ppid])
					
					format_memory_usage = self.method(:format_memory_usage).curry
					shared_memory_usage = 0
					private_memory_usage = 0
					
					proportional = true
					
					summary.each do |pid, general|
						terminal.print_line(:pid, pid, :reset, " ", :command, general[:command])
						
						terminal.print(:key, "Processor Usage: ".rjust(20), :reset)
						format_processor_utilization(general.processor_utilization, terminal)
						terminal.print_line
						
						if memory = general.memory
							shared_memory_usage += memory.proportional_size
							private_memory_usage += memory.unique_size
							
							terminal.print_line(
								:key, "Memory (PSS): ".rjust(20), :reset,
								format_memory_usage[memory.proportional_size]
							)
							
							terminal.print_line(
								:key, "Private (USS): ".rjust(20), :reset,
								format_memory_usage[memory.unique_size]
							)
						else
							shared_memory_usage += general.rss
							proportional = false
							
							terminal.print_line(
								:key, "Memory (RSS): ".rjust(20), :reset,
								format_memory_usage[general.rss]
							)
						end
					end
					
					terminal.print_line("Summary")
					
					if proportional
						terminal.print_line(
							:key, "Memory (PSS): ".rjust(20), :reset,
							format_memory_usage[shared_memory_usage]
						)
						
						terminal.print_line(
							:key, "Memory (USS): ".rjust(20), :reset,
							format_memory_usage[private_memory_usage]
						)
					else
						terminal.print_line(
							:key, "Memory (RSS): ".rjust(20), :reset,
							format_memory_usage[memory_usage]
						)
					end
					
					terminal.print_line(
						:key, "Memory (Total): ".rjust(20), :reset,
						format_memory_usage[shared_memory_usage + private_memory_usage]
					)
				end
			end
		end
	end
end
